--[[	Jump Game v8 -- The Musical
		Copyright 2009, 2010 Robin Wellner

		This work 'as-is' I provide.
		No warranty, express or implied.
		I've done my best,
		to debug and test.
		Liability for damages denied.

		Permission is granted hereby,
		to copy, share, and modify.
		Use as is fit,
		free or for profit.
		On this notice these rights rely.

]]
--[[	Clock sound by "Stephen"
		(http://free-loops.com/download-free-loop-2858.html)
		License: Creative Commons "Audio: Public1"?
		(I will treat it as CC-BY-NC-ND until I figure
		out what they mean by above license.)
]]

credits = {	"Jump Game v8 (the musical)",
			"Game design: Robin Wellner",
			'Clock sound: "Stephen"',
			"Music: (all CC-BY)",
			"The Heroes - Greenleo",
			"Jump And Run - bart@opengameart.org",
			"Suggestions: the Love community",
			"If you think you should be on this list,",
			"please contact me.", }

require "clouds.lua"
require "jgui.lua"

function Image(imgfile)
	return love.graphics.newImage("img/"..imgfile..".png")
end
function Sound(sndfile, streaming)
	local s = love.audio.newSource("snd/"..sndfile..".ogg", streaming or 'static')
	s:setLooping(true)
	return s
end

function love.load()
	Cloud:load()
	img =	{
			player = Image"player",
			badguy = Image"badguy",
			star = Image"star5",
			playerdead = Image"playerdead",
			badguydead = Image"badguydead",
			stardead = Image"star5dead",
			starcool = Image"star5cool",
			starfloat = Image"star5float",
			platform = Image"platform",
			wall = Image"wall",
			strongwall = Image"strongwall",
			menu = Image"menu5",
			powerup = Image"powerup",
			help = Image"help2",
			}
	snd =	{
			clock = Sound"clock",
			menu = Sound("The_Heroes", 'stream'),
			game = Sound("Random", 'stream'),
			}
	player = {x = 0, y = 0, vx = 0, vx_change = 0, vy = 0, vy_change = 0, dead=false, dx=0, dy=0, update=aifuncs.sway,
				stars = {n=0}, has_powerup=false, ddy=0, jumppower=3}
	state = {stage = 1, menu=true, menu_main=true, menu_index=1, menu_items = {"Continue", "Next Level", "Restart", "Load Level...", "Exit Game"},
				main_items = {"Select a Level...", "Level Editor...", "Help", "Credits", "Exit Game"},
				curlevel = nil}
	powerup = {x=0,y=0,hh=16,hw=16}
	unit2pixel = 20 -- pixels/unit
	pixel2unit = 1/unit2pixel --units/pixel
	clearstate()
	font = love.graphics.newFont("eurof55.ttf", 26)
	love.graphics.setFont(font)
	love.graphics.setLineWidth(6)
	phase = 0
	state.menu = true
	state.menu_main = true
	player.update=aifuncs.sway

	screen = jScreen:new()
	screen.textfield = jText:new{screen = screen}

	love.audio.play(snd.menu)
end

function loadlevel(file)
	clearstate()
	love.filesystem.load(file)()
	for i=1,#platform_list do
		platform_list[i].hw = 18.5
		platform_list[i].hh = 4.5
	end
	for i=1,#badguy_list do
		badguy_list[i].hw = 18
		badguy_list[i].hh = 18
	end
	for i=1,#wall_list do
		wall_list[i].hw = 32
		wall_list[i].hh = 32
	end
	for i=1,#star_list do
		star_list[i].hw = 13
		star_list[i].hh = 13
	end
	state.curlevel = file
	
	love.audio.rewind(snd.game)
	love.audio.play(snd.game)
end

function AddPlatform(arg)
	table.insert(platform_list, arg)
end
function AddBadGuy(arg)
	table.insert(badguy_list, arg)
end
function AddWall(arg)
	table.insert(wall_list, arg)
end
function AddStar(arg)
	table.insert(star_list, arg)
end

function clearstate()
	player.x = 0
	player.y = 0
	player.vx = 0
	player.vy = 0
	player.hw = 36/2
	player.hh = 36/2
	player.dead = false
	player.dx = 0
	player.dy = 0
	player.jumppower = 3
	player.update = aifuncs.nop
	player.stars = {n=0}
	player.has_powerup = false
	state.stage = 1
	state.menu = false
	state.menu_main = false
	state.menu_index = 1
	powerup.x = -10
	badguy_list = {}
	star_list = {}
	wall_list = {}
	platform_list = {}
	function win_condition () end
	function continue_condition() return #player.stars >= 7 end
	next_level = nil
end

function dist(a, b)
	return math.sqrt((a.x-b.x)*(a.x-b.x)+(a.y-b.y)*(a.y-b.y))
end

tdt = 0
function love.update(dt)
	if transitioning then
		transitioning = transitioning + dt*2
		if transitioning >= 1 then
			love.audio.stop(snd.menu)
			snd.game:setVolume(1)
			transitioning = nil
		else
			snd.menu:setVolume(1 - transitioning)
			snd.game:setVolume(transitioning)
		end
	end
	tdt = tdt + dt
	while tdt > .02 do
		tdt = tdt - .02
		realupdate(.02)
	end
	love.timer.sleep(state.menu and 40 or 20)
end

function realupdate(dt)
	Cloud:update(dt)
	player:update(dt)
	for i,badguy in ipairs(badguy_list) do
		badguy:update(dt)
		if badguy.dead then
			badguy.update = aifuncs.ascend
			--if gety(badguy) < -20 then table.remove(badguy_list, i) end
		elseif dist(player, badguy) < 36*pixel2unit and not player.dead then
			badguy.dead = true
			local n = #player.stars
			if n > 0 then
				player.stars[n].dead = true
				table.remove(player.stars, n)
				--table.sort(player.stars, function(x,y) return math.random(2)>1 end)
				for i=n-1,1,-1 do
					j = math.random(n-1)
					player.stars[i].followpos, player.stars[j].followpos =
						player.stars[j].followpos, player.stars[i].followpos
				end

			else
				player.dead = true
			end
		end
	end
	for i,star in ipairs(star_list) do
		star:update(dt)
		if star.dead then
			star.update = aifuncs.ascend
			--if gety(star) < -20 then table.remove(star_list, i) end
		elseif dist(player, star) < 31*pixel2unit and not player.dead then
			star.update = aifuncs.follow
			if not star.followpos then
				table.insert(player.stars, star)
				star.followpos = #player.stars
			end
		end
	end
	snd.game:setPitch(1+.01*#player.stars + (player.dead and -.06 or state.stage == 2 and .07 or player.has_powerup and .05 or 0))
	if dist(player, powerup) < 31*pixel2unit then
		player.has_powerup = true
		powerup.y = -10
	end
	if player.dead then
		player.update = aifuncs.ascend
	end
	if player.vx_change > 0.05 then
		player.vx_change = player.vx_change - .2
		player.vx = player.vx + .2
	elseif player.vx_change < -0.05 then
		player.vx_change = player.vx_change + .2
		player.vx = player.vx - .2
	end
	if player.vy_change > 0.02 then
		player.vy_change = player.vy_change - .2
		player.vy = player.vy + .2
	elseif player.vy_change < 0 then
		player.vy_change = player.vy_change + .5
		player.vy = player.vy - .5
	end
	if not player.dead then
		if not player.floatmode then
			player.dy = player.dy - 10*dt
			if player.dy < -20 then player.dy = -20 end
		end
		local r = player.x + player.dx*dt - player.vx
		if r < 0 then
			if player.vx > 0 then
				player.vx_change = -20
			end
		elseif r > (love.graphics.getWidth()-36)*pixel2unit then
			if continue_condition() then
				--player.vx = player.vx + 20
				player.vx_change = 20
			end
		else
			coll, str = collidewall()
			if not coll or (player.floatmode and not str) then
				player.x = player.x + player.dx*dt
			end
		end
		if collideplatform() and not player.floatmode then
			if player.dy < -10 then
				player.dy = -player.dy*.4
			else
				player.dy = 0
				player.jumppower = 3
			end
		else
			local _t = player.y + player.dy*dt
			if _t > 0 then
				local r = _t - player.vy
				if math.abs(player.vy_change) < 0.1 then
					if r < 0 and player.vy > 0 then
						player.vy_change = -20
					elseif r > (love.graphics.getHeight()-36)*pixel2unit then
						player.vy_change = 20
					end
				end
				player.y = _t
			else
				player.y = 0
				if player.dy < -10 then
					player.dy = -player.dy*.4
				else
					player.dy = 0
					player.jumppower = 3
				end
			end
		end
		if not leftdn and not rightdn then
			player.dx = player.dx * .1^dt
		else
			if math.abs(player.dx) < 7 then
				player.dx = player.dx + sign(player.dx)*.1
			end
		end
		if player.floatmode then
			player.jumppower = player.jumppower - dt
			if player.jumppower < 0 then
				player.floatmode = false
				love.audio.stop(snd.clock)
			end
			if not updn and not downdn then
				player.dy = player.dy * .1^dt
			else
				if math.abs(player.dy) < 7 then
					player.dy = player.dy + sign(player.dy)*.1
				end
			end
		end
	end
	phase = phase + .7*dt
	if win_condition() and not state.menu then
		--player.x > 200 for level 1
		state.stage = 2
		state.menu = true
		state.menu_index = 2
	end
end

function collideplatform()
	if player.dy > 0 then return false end
	for i,platform in ipairs(platform_list) do
		if player.x > platform.x - 1.5 and player.x < platform.x + 1.5 then
			if player.y > platform.y + .75 and player.y < platform.y + 1.05 then
				return true
			end
		end
	end
	return false
end

function sign(x)
	return x>0 and 1 or x<0 and -1 or 0
end

function collidewall()
	for i,wall in ipairs(wall_list) do
		if player.y > wall.y - 2 and player.y < wall.y + 2 then
			if player.x >= wall.x - 2 and player.x <= wall.x + 2 then
				if sign(player.dx) == sign(wall.x - player.x) then
					return true, wall.strong
				end
			end
		end
	end
	return false
end

function genlevellist(lvl)
	local _levels = lvl or {}
	for i,file in ipairs(love.filesystem.enumerate'') do
		if file:sub(-6) == '.level' then
			table.insert(_levels,file:sub(1,-7))
		end
	end
	table.insert(_levels,'Cancel')
	state.levellist = _levels
	state.menu = 'loadlevel'
	state.menu_index = 1
end

function love.keypressed(key)
	if state.menu==true then
		if state.menu_main then
			local nmenu = #state.main_items
			if key == 'up' then
				state.menu_index = ((state.menu_index - 2) % nmenu) + 1
				if state.menu_index == 2 then state.menu_index = 1 end
			elseif key == 'down' then
				state.menu_index = (state.menu_index % nmenu) + 1
				if state.menu_index == 2 then state.menu_index = 3 end
			elseif key == 'return' then
				if state.menu_index == 1 then
					genlevellist()
				elseif state.menu_index == 2 then
					--start level editor
					genlevellist({'New'})
					state.menu = 'loadeditor'
				elseif state.menu_index == 3 then
					--help?
					state.menu = 'help'
				elseif state.menu_index == 4 then
					--credits?
					state.menu = 'credits'
				elseif state.menu_index == 5 then
					exit()
				end
			end
		else
			local nmenu = #state.menu_items
			if key == 'up' then
				state.menu_index = ((state.menu_index - 2) % nmenu) + 1
				if state.stage == 2 and state.menu_index == 1 then
					state.menu_index = nmenu
				end
			elseif key == 'down' then
				state.menu_index = (state.menu_index % nmenu) + 1
				if state.stage == 2 and state.menu_index == 1 then
					state.menu_index = 2
				end
			elseif key == 'return' then
				if state.menu_index == 1 then
					state.menu = false
				elseif state.menu_index == 2 then
					if next_level and state.stage == 2 then
						loadlevel(next_level)
					end
				elseif state.menu_index == 3 then
					if state.curlevel then
						loadlevel(state.curlevel)
					end
				elseif state.menu_index == 4 then
					genlevellist()
				else
					exit()
				end
			end
		end
	elseif state.menu == 'help' then
		state.menu=true
	elseif state.menu == 'credits' then
		if key == 'escape' then
			state.menu=true
		end
	elseif state.menu == 'loadlevel' then
		local nmenu = #state.levellist
		if key == 'up' then
			state.menu_index = ((state.menu_index - 2) % nmenu) + 1
		elseif key == 'down' then
			state.menu_index = (state.menu_index % nmenu) + 1
		elseif key == 'return' then
			local newlevel = state.levellist[state.menu_index]
			if newlevel == 'Cancel' then
				state.menu = true
				state.menu_index = 1
			else
				--love.audio.stop(snd.menu)
				snd.game:setVolume(0)
				transitioning = 0
				loadlevel(newlevel..'.level')
			end
		end
	elseif state.menu == 'loadeditor' then
		local nmenu = #state.levellist
		if key == 'up' then
			state.menu_index = ((state.menu_index - 2) % nmenu) + 1
		elseif key == 'down' then
			state.menu_index = (state.menu_index % nmenu) + 1
		elseif key == 'return' then
			local newlevel = state.levellist[state.menu_index]
			if newlevel == 'Cancel' then
				state.menu = true
				state.menu_index = 2
			elseif newlevel == 'New' then
				state.menu = 'editor'
				clearstate()
				state.curlevel = nil
			else
				loadlevel(newlevel..'.level')
				state.menu = 'editor'
			end
		end
	elseif state.menu == 'editor' then
		screen:keypressed(key)
	else
		if not player.dead then
			if player.floatmode then
				if key == 'up' or key == 'w' then
					player.dy = 1
					updn = true
				elseif key == 'left' or key == 'a' then
					player.dx = -1
					leftdn = true
				elseif key == 'right' or key == 'd' then
					player.dx = 1
					rightdn = true
				elseif key == 'down' or key == 's' then
					player.dy = -1
					downdn = true
				end
			else
				if key == 'up' or key == 'w' then
					if player.dy == 0 then
						player.dy = 10
					elseif player.has_powerup and player.jumppower > 0 then
						player.dy = 10
						player.jumppower = player.jumppower - 1
					end
				elseif key == 'left' or key == 'a' then
					player.dx = -1
					leftdn = true
				elseif key == 'right' or key == 'd' then
					player.dx = 1
					rightdn = true
				elseif key == 'down' or key == 's' then
					player.dx = 0
				elseif key == ' ' and player.has_powerup and player.jumppower > 0 then
					player.floatmode = true
					love.audio.play(snd.clock, 0)
				else
				end
			end
		end
		if key == 'escape' then
			state.menu = true
		end
	end
end

function love.keyreleased(key)
	if key == 'left' or key == 'a' then
		leftdn = false
	elseif key == 'right' or key == 'd' then
		rightdn = false
	elseif key == 'up' or key == 'w' then
		updn = false
	elseif key == 'down' or key == 's' then
		downdn = false
	elseif key == ' ' then
		player.floatmode = false
		love.audio.stop(snd.clock)
	end
end

function getx(self)
	return 18+self.x*unit2pixel - player.vx*unit2pixel - self.hw
end

function gety(self)
	return love.graphics.getHeight() - 18 - self.y*unit2pixel + player.vy*unit2pixel - self.hh
end

function love.draw()
	Cloud:drawall()
	local img = img
	for i,platform in ipairs(platform_list) do
		love.graphics.draw(img.platform, getx(platform), gety(platform))
	end
	for i,wall in ipairs(wall_list) do
		local _w
		if wall.strong then
			_w = img.strongwall
		else
			_w = img.wall
		end
		love.graphics.draw(_w, getx(wall), gety(wall))
	end
	if player.dead or player.floatmode then
		love.graphics.draw(img.playerdead, getx(player), gety(player))
	else
		love.graphics.draw(img.player, getx(player), gety(player))
		--love.graphics.draw(img.player, getx(player), gety(player), 0, 1, 1-player.ddy*.05)
	end
	love.graphics.draw(img.powerup, getx(powerup), gety(powerup))
	for i,badguy in ipairs(badguy_list) do
		if badguy.dead then
			love.graphics.draw(img.badguydead, getx(badguy), gety(badguy))
		else
			love.graphics.draw(img.badguy, getx(badguy), gety(badguy))
		end
	end
	local img_star
	for i,star in ipairs(star_list) do
		if star.dead then
			img_star = img.stardead
		elseif player.has_powerup and star.followpos  then
			if player.floatmode then
				img_star = img.starfloat
			else
				img_star = img.starcool
			end
		else
			img_star = img.star
		end
		love.graphics.draw(img_star, getx(star), gety(star))
	end
	if state.menu == 'help' then
		love.graphics.draw(img.help, 400-img.help:getWidth()*.5, 300-img.help:getHeight()*.5)
	elseif state.menu == 'editor' then
		screen:draw()
	elseif state.menu then
		love.graphics.draw(img.menu, 400-img.menu:getWidth()*.5, 300-img.menu:getHeight()*.5)
		local mnux = 180
		local mnuy = 120
		if state.menu == 'loadlevel' then
			for i,item in ipairs(state.levellist) do
				--love.graphics.setColor(0, 0, 0)
				local mnuh = i * 40 + mnuy
				if i == state.menu_index then
					love.graphics.line(mnux, 5 + mnuh + font:getHeight(item), mnux + font:getWidth(item), 5 + mnuh + font:getHeight(item))
					--love.graphics.setColor(0, 0, 100)
				end
				love.graphics.print(item, mnux, mnuh)
			end
		elseif state.menu == 'credits' then
			for i,item in ipairs(credits) do
				local mnuh = i * 40 + mnuy
				love.graphics.print(item, mnux, mnuh)
			end
		elseif state.menu_main then
			for i,item in ipairs(state.main_items) do
				--love.graphics.setColor(0, 0, 0)
				local mnuh = i * 40 + mnuy
				if i == state.menu_index then
					love.graphics.line(mnux, 5 + mnuh + font:getHeight(item), mnux + font:getWidth(item), 5 + mnuh + font:getHeight(item))
					--love.graphics.setColor(0, 0, 100)
				end
				if i == 2 then
					love.graphics.setColor(160, 160, 160)
				else
					love.graphics.setColor(255, 255, 255)
				end
				love.graphics.print(item, mnux, mnuh)
				love.graphics.setColor(255, 255, 255)
			end
		else
			if state.stage == 2 then
				love.graphics.print("Congratulations, you have made it!", mnux, mnuy+40)
			end
			for i,item in ipairs(state.menu_items) do
				if i > 1 or state.stage == 1 then
					--love.graphics.setColor(0, 0, 0)
					local mnuh = i * 40 + mnuy
					if i == state.menu_index then
						love.graphics.line(mnux, 5 + mnuh + font:getHeight(item), mnux + font:getWidth(item), 5 + mnuh + font:getHeight(item))
						--love.graphics.setColor(0, 0, 100)
					end
					love.graphics.print(item, mnux, mnuh)
				end
			end
		end
	end
end

function exit()
	love.event.push'q'
end


aifuncs = {}
function aifuncs:nop(dt)
end
function aifuncs:ascend(dt)
	self.y = self.y + (.5+self.y/10)*dt
end
function aifuncs:sway(dt)
	if not self.m then
		if self.x <= 0.1 then
			self.dy = 5
			self.dx = 0
			self.m = 1
		else
			self.dx = self.dx - .1
		end
	elseif self.m == 1 then
		if self.y == 0 then
			self.dx = 5
			self.m = 2
		end
	elseif self.m == 2 then
		if self.x >= 4 then
			self.dy = 5
			self.dx = 0
			self.m = 3
		else
			self.dx = self.dx + .1
		end
	elseif self.m == 3 then
		if self.y == 0 then
			self.dx = -5
			self.m = nil
		end
	end
end
function aifuncs.bounce(xmin, xmax)
	return function (self, dt)
		if self.m then
			self.x=self.x-dt
			if self.x < xmin then self.m = false end
		else
			self.x=self.x+dt
			if self.x > xmax then self.m = true end
		end
	end
end
function aifuncs:follow(dt)
	local theta = 0 ; local gotox = 0 ; local gotoy = 0
	if self.followpos < 10 then
		theta = self.followpos*.7+phase
		gotox = player.x + 60*pixel2unit * math.cos(theta)
		gotoy = player.y + 60*pixel2unit * math.sin(theta)
	elseif self.followpos < 26 then
		theta = self.followpos*.4+phase
		gotox = player.x + 120*pixel2unit * math.cos(theta)
		gotoy = player.y + 120*pixel2unit * math.sin(theta)
	elseif self.followpos < 47 then
		theta = self.followpos*.3+phase
		gotox = player.x + 180*pixel2unit * math.cos(theta)
		gotoy = player.y + 180*pixel2unit * math.sin(theta)
	elseif self.followpos < 88 then
		theta = self.followpos*.15+phase
		gotox = player.x + 240*pixel2unit * math.cos(theta)
		gotoy = player.y + 240*pixel2unit * math.sin(theta)
	else
		theta = self.followpos*.12+phase
		gotox = player.x + 300*pixel2unit * math.cos(theta)
		gotoy = player.y + 300*pixel2unit * math.sin(theta)
	end
	if gotox - self.x > 0 then
		self.x = self.x + 2*dt + (gotox - self.x)*pixel2unit*.2
	elseif gotox - self.x < 0 then
		self.x = self.x - 2*dt + (gotox - self.x)*pixel2unit*.2
	end
	if gotoy - self.y > 0 then
		self.y = self.y + 2*dt + (gotoy - self.y)*pixel2unit*.2
	elseif gotoy - self.y < 0 then
		self.y = self.y - 2*dt + (gotoy - self.y)*pixel2unit*.2
	end
end
function aifuncs.guard(xmin, xmax, super)
	local width, low, high, speed
	if super then
		width = 5
		low = 3
		high = 30
		speed = 9
	else
		width = 2.5
		low = 1.5
		high = 8.5
		speed = 3
	end
	return function (self, dt)
		if player.y > self.y - low and player.y < self.y + high
			and player.x > xmin - width and player.x < xmax + width
				then
			local dx = player.x - self.x
			local dt2 = speed*dt
			if dx > dt2 then
				dx = dt2
			elseif dx < -dt2 then
				dx = -dt2
			end
			self.x = self.x + dx
		end
	end
end
function aifuncs:block(dt)
	if not self.dy then self.dy = 0 end
	if player.y > self.y and player.y < 10
		and player.x > self.x - 10 and player.x < self.x + 10
			then
		self.dy = math.min((player.y - self.y)*3, 6)
	else
		self.dy = self.dy - .1
	end
	if self.y+self.dy*dt > 0 then
		self.y = self.y + self.dy*dt
	end
end